unit trayicfm;

// Copyright  1999 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

// Re-usable tray icon form.
// Revised 3/8/1999 to correctly handle WM_(QUERY)ENDSESSION
// Revised 3/15/1999 to make animation optional (Animate property)
// Revised 3/16/1999 to automatically restore tray icon after
//    an Explorer crash

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, StdCtrls;

type
  TTrayIconForm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action:
      TCloseAction);
  private
    { Private declarations }
    fTrayIcon : hIcon;
    fTip      : String;
    fID       : Cardinal;
    fAnimate  : Boolean;
    fReady    : Boolean;
    fTruClose : Boolean;
    fShutdown : Boolean;
    fTaskMsg  : DWORD;
    fStopEv   : THandle;
    fTrayEv   : THandle;
    procedure SetTrayIcon(value: hIcon);
    procedure SetTip(const value: String);
    function  ShellNotify(Msg, Flags: Integer): Boolean;
  protected
    procedure wmSyscommand(VAR Msg: TWMSysCommand);
      message WM_SYSCOMMAND;
    procedure wmEndSession(VAR Msg: TWMEndSession);
      message WM_ENDSESSION;
    procedure wmQueryEndSession(VAR Msg: TWMQueryEndSession);
      message WM_QUERYENDSESSION;
    procedure AnimateRect(IsHide : Boolean);
    procedure TrulyClose;
    function OpenWhat : TForm; virtual;
  public
    { Public declarations }
    function CloseQuery : Boolean; override;
    procedure DefaultHandler(var Message); override;
    procedure SetupTrayIcon(vID: Integer; vTrayIcon: hIcon;
      const vTip : String);
    property TrayIcon     : hIcon Write SetTrayIcon;
    property Tip          : String Write SetTip;
    property Animate      : Boolean read fAnimate write fAnimate;
  end;

implementation
USES ShellApi, mmSystem, Registry;
{$R *.DFM}

function ShellVersionIs4_70 : boolean;
type
  PDllVersionInfo = ^TDllVersionInfo;
  TDllVersionInfo = record
    cbSize          : DWORD;
    dwMajorVersion  : DWORD;
    dwMinorVersion  : DWORD;
    dwBuildNumber   : DWORD;
    dwPlatformID    : DWORD;
  end;
  DllGetVersionType = function (pdvi : PDllVersionInfo) :
    HRESULT; stdcall;
VAR
  LibH : hModule;
  DllGetVersion  : DllGetVersionType;
  DVI  : TDllVersionInfo;
begin
  Result := False;
  LibH := LoadLibrary('Shell32.dll');
  IF LibH <= 32 THEN Exit;
  try
    @DllGetVersion := GetProcAddress(LibH, 'DllGetVersion');
    IF @DllGetVersion = nil THEN Exit;
    FillChar(DVI, SizeOf(DVI), 0);
    DVI.cbSize := SizeOf(DVI);
    IF DllGetVersion(@DVI) = NOERROR THEN
      begin
        IF DVI.dwMajorVersion > 4 THEN Result := True
        ELSE IF (DVI.dwMajorVersion = 4) AND
          (DVI.dwMinorVersion >= 70) THEN Result := True;
      end;
  finally
    FreeLibrary(LibH);
  end;
end;

function ProcHandleForTray : THandle;
VAR
  Trayhan : HWnd;
  TrayPID : DWORD;
begin
  Result := 0;
  TrayHan := FindWindowEx(GetDesktopWindow, 0, 'Shell_TrayWnd', nil);
  IF TrayHan = 0 THEN Exit;
  GetWindowThreadProcessID(TrayHan, @TrayPID);
  IF TrayPID = 0 THEN Exit;
  Result := OpenProcess(STANDARD_RIGHTS_REQUIRED OR SYNCHRONIZE,
    False, TrayPID);
end;

function ReThread(Who : TTrayIconForm) : DWORD; export; stdcall;
VAR
  N : Integer;
  fhans : ARRAY[0..1] OF THandle;
begin
  Result := 0;
  fhans[0] := Who.fStopEv;
  fhans[1] := Who.fTrayEv;
  WHILE TRUE DO
    begin
      // Wait for the Stop event (sent when program ends) or for
      // the tray process handle. If the latter, wait 'til the
      // tray is back and then re-show our icon.
      CASE WaitForMultipleObjects(2, @fhans, False, INFINITE) OF
        WAIT_OBJECT_0 + 1 : begin
          CloseHandle(fhans[1]);
          fhans[1] := ProcHandleForTray;
          N := 0;
          WHILE (fHans[1] = 0) AND (N < 10) DO
            begin
              IF WaitForSingleObject(fHans[0], 250) <> WAIT_TIMEOUT
                THEN Break;
              fHans[1] := ProcHandleForTray;
              Inc(N);
            end;
          Who.ShellNotify(NIM_ADD, NIF_TIP OR NIF_MESSAGE OR
            NIF_ICON);
        end;
        ELSE Break;
      END;
    end;
end;

procedure TTrayIconForm.FormCreate(Sender: TObject);
VAR thread_id : DWORD;
begin
  fTaskMsg := RegisterWindowMessage('TaskbarCreated');
  // Hide the taskbar button
  ShowWindow(Application.Handle, SW_HIDE);
  // Don't show the main form to start with
  Application.ShowMainForm := False;
  fTruClose := False;
  fShutdown := False;
  fAnimate  := False;
  IF NOT ShellVersionIs4_70 THEN
    begin
      fStopEv := CreateEvent(nil, False, False,
        PChar(Self.Name + ' TrayIconForm Stop Event'));
      fTrayEv := ProcHandleForTray;
      CreateThread(nil, 4096, @ReThread, Self, 0, thread_id);
    end;
end;

procedure TTrayIconForm.FormHide(Sender: TObject);
begin
  IF (NOT fTruClose) AND Animate THEN
    // hide occurs during close; don't animate!
    begin
      // If system sound for minimizing is defined, play it
      PlaySound('RestoreDown', 0, SND_ALIAS OR SND_ASYNC OR
        SND_NODEFAULT);
      // Implode to taskbar
      AnimateRect(True);
    end;
end;

procedure TTrayIconForm.FormShow(Sender: TObject);
begin
  IF Animate THEN
    begin
      // If system sound for restoring is defined, play it
      PlaySound('RestoreUp', 0, SND_ALIAS OR SND_ASYNC OR SND_NODEFAULT);
      // Explode from taskbar
      AnimateRect(False);
    end;
end;

procedure TTrayIconForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  ShellNotify(NIM_DELETE, 0);
  PulseEvent(fStopEv);
  CloseHandle(fStopEv);
  CloseHandle(fTrayEv);
end;

procedure TTrayIconForm.SetTrayIcon(value: hIcon);
begin
  IF NOT fReady THEN Exit;
  IF Value = fTrayIcon THEN Exit;
  fTrayIcon := Value;
  ShellNotify(NIM_MODIFY, NIF_ICON)
end;

procedure TTrayIconForm.SetTip(const value: String);
begin
  IF NOT fReady THEN Exit;
  IF Value = fTip THEN Exit;
  fTip := Value;
  ShellNotify(NIM_MODIFY, NIF_TIP)
end;

function TTrayIconForm.ShellNotify(Msg, Flags: Integer) : Boolean;
VAR tnd : TNotifyIconData;
BEGIN
  Result := False;
  IF fID = 0 THEN Exit;
  FillChar(tnd, SizeOf(tnd), 0);
  WITH tnd DO
    BEGIN
      Wnd := Self.Handle;
      uFlags := Flags;
      IF Msg = NIM_ADD THEN
        uFlags := uFlags OR NIF_MESSAGE OR NIF_ICON OR NIF_TIP;
      cbSize := SizeOf(tnd);
      uID := fID;
      uCallbackMessage:= WM_USER + fID;
      IF fTrayIcon <> 0 THEN hIcon:= fTrayIcon
      ELSE hIcon := LoadIcon(hInstance, IDI_WINLOGO);
      StrLCopy(szTip, PChar(fTip), 64);
      Result := Shell_NotifyIcon(msg, @tnd);
      IF Result THEN DestroyIcon(fTrayIcon);
    END;
END;

procedure TTrayIconForm.wmSyscommand(VAR Msg: TWMSysCommand);
// On system-menu Close, just minimize
begin
  IF Msg.CmdType AND $FFF0 = SC_MINIMIZE THEN
    BEGIN
      Close;
      Msg.Result := 0;
    END
  ELSE Inherited;
end;

procedure TTrayIconForm.wmEndSession(VAR Msg: TWMEndSession);
begin
  IF Msg.EndSession THEN
    TrulyClose
  else
    begin
      fTruClose := False;
      fShutDown := False;
    end;
  Msg.Result := 0;
end;

procedure TTrayIconForm.wmQueryEndSession(VAR Msg: TWMQueryEndSession);
begin
  fTruClose := True;
  IF CloseQuery THEN
    begin
      Msg.Result := 1;
      fTruClose := True;
      fShutdown := True;
    end
  ELSE
    begin
      Msg.Result := 0;
      ftruClose := False;
      fShutdown := False;
    end;
end;

procedure TTrayIconForm.AnimateRect(IsHide : Boolean);
VAR
  STWH, TNWH : Integer;
  GotR2      : Boolean;
  R1, R2     : TRect;
  tabd       : TAppBarData;
begin
  WITH TRegIniFile.Create('Control Panel\Desktop\WindowMetrics') DO
  try
    IF ReadString('', 'MinAnimate', '1') <> '1' THEN Exit;
  finally
    Free;
  end;
  GotR2 := False;
  // Get the system tray's bounding rectangle
  STWH := FindWindow('Shell_TrayWnd', nil);
  IF STWH <> 0 THEN
    begin
     TNWH := FindWindowEx(STWH, 0, 'TrayNotifyWnd', nil);
     IF TNWH <> 0 THEN
       begin
         GetWindowRect(TNWH, R2);
         GotR2 := True;
       end;
    end;
  // If that fails, invent a rectangle in the correct corner
  IF NOT GotR2 THEN
    begin
      FillChar(tabd, SizeOf(tabd), 0);
      tabd.cbSize := SizeOf(tabd);
      IF SHAppBarMessage(ABM_GETTASKBARPOS, Tabd)=0 THEN Exit;
      WITH Screen, tabd.rc DO
        IF (top > 0) OR (left > 0) THEN
          R2 := Rect(Width-32, Height-32, Width, Height)
        ELSE IF (bottom < Height) THEN
          R2 := Rect(Width-32, 0, Width, 32)
        ELSE IF (right < Width) THEN
          R2 := Rect(0, Height-32, 32, Height)
        ELSE Exit;
    end;
  R1 := BoundsRect;
  IF IsHide THEN
    DrawAnimatedRects(Handle, IDANI_CAPTION, R1, R2)
  ELSE
    DrawAnimatedRects(Handle, IDANI_CAPTION, R2, R1);
end;

procedure TTrayIconForm.TrulyClose;
begin
  fTruClose := True;
  Close;
end;

function TTrayIconForm.OpenWhat : TForm;
// override if clicking icon should open different window
begin
  Result := Self
end;

function TTrayIconForm.CloseQuery: Boolean;
begin
  IF fShutDown THEN
    Result := True
  ELSE IF fTruClose THEN
    begin
      Result := Inherited CloseQuery;
      IF NOT Result THEN
        fTruClose := False;
    end
  ELSE
    begin
      Result := False;
      Hide;
    end;
end;

procedure TTrayIconForm.DefaultHandler(var Message);
VAR
  CP : TPoint;
  TF : TForm;
begin
  Inherited DefaultHandler(Message);
  WITH TMessage(Message) DO
    IF Msg = fTaskMsg THEN
      begin
        // Restore tray icon when Explorer crashes/restarts
        ShellNotify(NIM_ADD, NIF_TIP OR NIF_MESSAGE OR NIF_ICON);
      end
    ELSE IF (fID <> 0) AND
      (Msg = WM_USER + fID) AND
      (wParam = LongInt(fID)) THEN
      CASE lParam OF
        WM_LBUTTONDOWN : begin
          TF := OpenWhat;
          IF TF <> nil THEN
            begin
              IF NOT TF.Visible THEN TF.Show;
              SetForegroundWindow(TF.Handle);
            end;
        end;
        WM_RBUTTONDOWN : IF Assigned(PopupMenu) THEN
          BEGIN
            TF := OpenWhat;
            IF TF <> nil THEN
              SetForegroundWindow(TF.Handle);
            GetCursorPos(CP);
            PopupMenu.Popup(CP.X, CP.Y);
          END
        ELSE Close;
      END;
end;

procedure TTrayIconForm.SetupTrayIcon(vID: Integer;
  vTrayIcon: hIcon; const vTip : String);
BEGIN
  fID       := vID;
  fTrayIcon := vTrayIcon;
  fTip      := vTip;
  fReady    := True;
  ShellNotify(NIM_ADD, NIF_TIP OR NIF_MESSAGE OR NIF_ICON);
END;

end.
