unit tidyobj;

// Copyright  2001 by Ziff Davis Media, Inc.
// Written by Neil J. Rubenking

interface

uses Windows, Classes, SysUtils, comctrls, commctrl, tidyfuncs, graphics;
type
  PShared9x = ^TShared9x;
  TShared9x = record
    Buff : ARRAY[0..MAX_PATH] OF Char;
    LVI  : TLVItem;
    Pt   : TPoint;
  end;
  VirtualAllocExType = function (hProcess : THandle;
    lpAddress : Pointer; dwSize, flAllocationType : DWORD;
    flProtect : DWORD) : Pointer; stdcall;
  VirtualFreeExType = function (hProcess : THandle;
    lpAddress : Pointer; dwSize, dwFreeType : DWORD) : Pointer;
    stdcall;

  TLayout = class(TObject)
  private
    // Stored in stream
    fList     : TStringList;
    fName     : String;
    fWid      : Integer;
    fHig      : Integer;
    fNormIcon : Boolean;
    fPredef   : Integer; // -1 if not predef
    // Not stored
    fRemains     : TStringList;
    // Used in inter-process communication
    UnderNT      : Boolean;   // true only if under NT-family
    lvHProc      : THandle;   // its process handle (NT only)
    R_buf        : PChar;     // stringbuffer, treeitem, and
    R_lvi        : PLVItem;   //     memory space
    R_pt         : PPoint;    // xx
    L_buf        : ARRAY[0..MAX_PATH] OF Char; // local
    L_lvi        : TLVItem;   //     copies of the items in
    L_pt         : TPoint;    //     desktop's space
    Map9x        : THandle;   // handle to memory-mapping (9x)
    Share9x      : PShared9x; // pointer to mapped memory (9x)
    KernelH      : THandle;
    vVirtualAllocEx : VirtualAllocExType;
    vVirtualFreeEx  : VirtualFreeExType;
    // inter-process communication functions
    function GetHProc : THandle;
    function AllocShared : Boolean;
    procedure FromShared;
    procedure ToShared;
    procedure DoneShared;

    procedure SetHig(const Value: Integer);
    procedure SetName(const Value: String);
    procedure SetWid(const Value: Integer);
    function lvHan : HWnd;
    function StoreStr(const S: String; vList: TStringList) : String;
    procedure MakeLayout(TS : TStringList; TC : TCanvas;
      vLayout : Integer);
    procedure ToDesktop_internal(TC : TCanvas);
  protected
  public
    constructor Create; overload;
    constructor Create(const vName : String); overload;
    constructor Create(vWid, vHig : Integer); overload;
    destructor Destroy; override;
    procedure LoadFromStream(Stream : TStream);
    procedure SaveToStream(Stream : TStream);
    class function IconsAreNormal : Boolean;
    class function PointToPointer(Pt : TPoint) : Pointer;
    class function PointerToPoint(P : Pointer) : TPoint;
    function Longest(TC : TCanvas) : Integer;
    function ReportDupes(H : HWnd; const S : String) : Boolean;
    function SetOrder(vList : TStrings) : boolean;
    procedure CheckIsPredef(cLayout : Integer; TC : TCanvas);
    procedure FromDesktop;
    procedure ReportRemains(H : HWnd; const S : String);
    procedure SetLayout(cLayout : Integer; TC : TCanvas);
    procedure SetListViewStyleTo(Normal : Boolean);
    procedure ToDesktop(TC : TCanvas);
    // properties
    property List     : TStringList read fList;
    property Name     : String read fName write SetName;
    property Wid      : Integer read fWid write SetWid;
    property Hig      : Integer read fHig write SetHig;
    property NormIcon : Boolean read fNormIcon write fNormIcon;
    property Predef   : Integer read fPredef write fPredef;
  published
end;

TLayoutList = class(TObject)
private
  fNamed : TStringList;
  fResol : TStringList;
  fOrder : TLayout;
protected
public
  constructor Create;
  destructor Destroy; override;
  procedure SaveToStream(Stream : TStream);
  procedure LoadFromStream(Stream : TStream);
  procedure SaveToFile(const S : String);
  procedure LoadFromFile(const S : String);
  procedure AddLayout(vTL : TLayout);
  procedure DelLayout(const Name : String); overload;
  procedure DelLayout(vWid, vHig : Integer); overload;
  function LayoutByName(const Name: String): TLayout;
  function LayoutByRes(vWid, vHig: Integer): TLayout;
  // properties
  property Named : TStringList read fNamed;
  property Order : TLayout read fOrder;
published

end;



implementation
uses Registry, messages;
{ TIconLoc }

const
  PSiz = SizeOf(TPoint);
  LSiz = SizeOf(TLvItem);
  Signature : DWORD = $59444954;
  DatVersion : DWord = $8000000A;
  cIcoAtomName = 'PC Magazine WinTidy Atom Name '+
    '{E4E3BFB4-7A38-4056-9688-2165CE91B157}';
  cMappingName = 'PC Magazine WinTidy data mapping '+
    '{1E4A2814-A67B-40D3-8E4E-C6AED03DD70E}';

{ TLayout }

constructor TLayout.Create;
begin
  Inherited Create;
  fName     := '';
  fWid      := -1;
  fHig      := -1;
  fPredef   := -1;
  fList     := TStringList.Create;
  fRemains  := TStringList.Create;
  fNormIcon := IconsAreNormal;
  UnderNT   := Win32Platform = VER_PLATFORM_WIN32_NT;
  lvHProc   := 0;
end;

constructor TLayout.Create(const vName: String);
begin
  Create;
  fName := vName;
end;

constructor TLayout.Create(vWid, vHig: Integer);
begin
  Create;
  fWid := vWid;
  fHig := vHig;
end;

destructor TLayout.Destroy;
begin
  fList.Free;
  fRemains.Free;
  inherited;
end;

procedure TLayout.LoadFromStream(Stream: TStream);
VAR
  Reader : TReader;
  Count  : Integer;
  N      : Integer;
  S      : String;
  X, Y   : Integer;
begin
  fList.Clear;
  Reader := TReader.Create(Stream, 4096);
  WITH Reader DO
    try
      fName     := ReadString;
      fWid      := ReadInteger;
      fHig      := ReadInteger;
      fNormIcon := ReadBoolean;
      fPredef   := ReadInteger;
      Count     := ReadInteger;
      FOR N := 0 TO Count-1 DO
        begin
          S := ReadString;
          X := ReadInteger;
          Y := ReadInteger;
          fList.AddObject(S, PointToPointer(Point(X,Y)));
        end;
    finally
      Free;
    end;
end;

procedure TLayout.SaveToStream(Stream: TStream);
VAR
  Writer : TWriter;
  N      : Integer;
begin
  Writer := TWriter.Create(Stream, 4096);
  WITH Writer DO
    try
      WriteString(fName);
      WriteInteger(fWid);
      WriteInteger(fHig);
      WriteBoolean(fNormIcon);
      WriteInteger(fPredef);
      WriteInteger(fList.Count);
      FOR N := 0 TO fList.Count-1 DO
        begin
          WriteString(fList[N]);
          WITH PointerToPoint(fList.Objects[N]) DO
            begin
              WriteInteger(X);
              WriteInteger(Y);
            end;
        end;
    finally
      Free;
    end;
end;

class function TLayout.IconsAreNormal: Boolean;
// The specified string is associated with an atom only when WinTidy
// has put the desktop in small icons mode. If GlobalFindAtom()
// returns 0, the atom doesn't exist and icons are normal
begin
  Result := GlobalFindAtom(cIcoAtomName) = 0;
end;

class function TLayout.PointToPointer(Pt : TPoint): Pointer;
// Used in storing a TPOINT location in the Objects array of a
// TStringList
begin
  Result := Pointer(Integer(PointToSmallPoint(Pt)))
end;

class function TLayout.PointerToPoint(P: Pointer): TPoint;
// Used in extracting a TPoint location from the Objects array of a
// TStringList
begin
  Result := SmallPointToPoint(TSmallPoint(Integer(P)));
end;

function TLayout.Longest(TC : TCanvas): Integer;
// Returns the length of the longest icon title, using the current
// system-defined icon font. For convenience, it takes an existing
// TCanvas object rather than creating/destroying one.
VAR
  N  : Integer;
  W  : Integer;
  IM : ICONMETRICS;
begin
  FillChar(IM, SizeOf(IM), 0);
  IM.cbSize := SizeOf(IM);
  SystemParametersInfo(SPI_GETICONMETRICS, SizeOf(IM), @IM, 0);
  TC.Font.Handle := CreateFontIndirect(im.lfFont);
  // Don't return a value under 32
  Result := 32;
  FOR N := 0 TO List.Count-1 DO
    begin
      W := TC.TextWidth(List[N]);
      IF W > Result THEN
        Result := W;
    end;
end;

function TLayout.ReportDupes(H : HWnd; const S : String) : Boolean;
// Report that the desktop contains duplicate items. This really should
// not be possible, given that WinTidy distinguishes between the
// user desktop, common desktop, and non-file items, and distinguishes
// different shortcut file types even though their extensions aren't
// reported on the desktop. But IF it happens, this method will report.
begin
  IF fRemains.Count > 10 THEN
    begin
      fRemains[9] := format('... and %d others', [fRemains.Count]);
      WHILE fRemains.Count > 10 DO
        fRemains.Delete(10);
    end;
  Result := (fRemains.Count > 0) AND (MessageBox(H,
    PChar(Format(sFmtDupes, [fRemains.Text, S])), 'WinTidy',
    MB_YESNO OR MB_ICONQUESTION) <> idYes);
end;

function TLayout.SetOrder(vList: TStrings) : boolean;
// Set the list of icons to match the order in the passed list. If
// either list contains icons not present in the other, returns
// TRUE, indicating a change.
VAR
  TS  : TStringList;
  N   : Integer;
  Idx : Integer;
begin
  Result := False;
  // Build a temporary list from the existing list, using the order
  // of the passed list
  TS := TStringList.Create;
  try
    FOR N := 0 TO vList.Count-1 DO
      begin
        Idx := fList.IndexOf(vList[N]);
        IF Idx = -1 THEN
          begin
            // An item not in our list - skip it and return true
            Result := True;
            Continue;
          end;
        TS.AddObject(fList[Idx], fList.Objects[Idx]);
        fList.Delete(Idx);
      end;
    IF fList.Count > 0 THEN
      begin
        // Some items in our list are not in the passed list. Add
        // them to the temporary list and return True
        Result := True;
        FOR N := 0 TO fList.Count-1 DO
          TS.AddObject(fList[N], fList.Objects[N]);
      end;
    fList.Assign(TS);
  finally
    TS.Clear;
  end;
end;

procedure TLayout.CheckIsPredef(cLayout: Integer; TC: TCanvas);
// Return TRUE only if the current icon locations precisely match
// the specified predefined layout
VAR
  Layout : TStringList;
  N      : Integer;
begin
  fPredef := -1;
  Layout := TStringList.Create;
  try
    MakeLayout(Layout, TC, cLayout);
    FOR N := 0 TO fList.Count-1 DO
      IF Integer(fList.Objects[N]) <> Integer(Layout.Objects[N]) THEN
        Exit;
    fPredef := cLayout;
  finally
    Layout.Free;
  end;
end;

procedure TLayout.FromDesktop;
VAR
  N     : Integer;
  Count : Integer;
  H     : HWnd;

  procedure GetOneIcon(N : Integer);
  VAR sN : String;
  begin
    WITH L_lvi DO
      begin
        mask       := LVIF_TEXT;
        iItem      := N;
        iSubItem   := 0;
        state      := 0;
        statemask  := 0;
        pszText    := R_buf;
        cchTextMax := MAX_PATH;
        iImage     := 0;
      end;
    ToShared;
    SendMessage(H, LVM_GETITEM, 0, Integer(R_lvi));
    SendMessage(H, LVM_GETITEMPOSITION, N, Integer(R_Pt));
    FromShared;
    sN := StoreStr(L_buf, fList);
    IF fList.IndexOf(sN) >= 0 THEN
      fRemains.Add(L_buf);
    // We process the icons from last to first, to aid in figuring
    // out their type. Thus each one is inserted at position 0
    // instead of being added at the end
    fList.InsertObject(0, sN,
      PointToPointer(L_pt));
  end;

begin
  H := LVHan;
  try
    fNormIcon := GlobalFindAtom(cIcoAtomName) = 0;
    IF UnderNT THEN lvHProc := GetHProc;
    IF NOT AllocShared THEN Exit;
    try
      fList.Clear;
      fRemains.Clear;
      Count := SendMessage(H, LVM_GETITEMCOUNT, 0, 0);
      FOR N := Count-1 DOWNTO 0 DO
        GetOneIcon(N);
    finally
      DoneShared;
    end;
  finally
    IF lvHProc <> 0 THEN CloseHandle(lvHProc);
  end;
end;

procedure TLayout.ReportRemains(H: HWnd; const S : String);
// If some icons on the desktop were not included in the just-restored
// layout, warn the user.
begin
  IF fRemains.Count = 0 THEN Exit;
  IF fRemains.Count > 10 THEN
    begin
      fRemains[10] := format('... and %d others', [fRemains.Count]);
      WHILE fRemains.Count > 11 DO
        fRemains.Delete(11);
    end;
  MessageBox(H, PChar(Format(sFmtRemains, [S, fRemains.Text])),
    'WinTidy', MB_OK OR MB_ICONINFORMATION);
end;

procedure TLayout.SetLayout(cLayout: Integer; TC : TCanvas);
VAR
  Layout : TStringList;
  N      : Integer;
const
  solu : ARRAY[Boolean] OF String =
  ('rename the icons with the longest titles',
   'use the Display Properties dialog to decrease icon spacing');
begin
  Layout := TStringList.Create;
  try
    MakeLayout(Layout, TC, cLayout);
    IF fList.Count > Layout.Count THEN
      begin
        MessageBox(0, PChar(Format(sFmtManyIcons, [fList.Count,
          Layout.Count, solu[IconsAreNormal]])), 'WinTidy',
          MB_OK OR MB_ICONSTOP);
      end
    ELSE
      begin
        FOR N := 0 TO fList.Count-1 DO
          WITH TSmallPoint(Layout.Objects[N]) DO
            fList.Objects[N] := PointToPointer(Point(X, Y));
      end;
  finally
    Layout.Free;
  end;
end;

procedure TLayout.SetListViewStyleTo(Normal : boolean);
VAR
  Styles : StyleStruct;
  H      : HWnd;
begin
  IF Normal = IconsAreNormal THEN Exit;
  IF (NOT Normal) AND ActiveDesktop THEN Exit;
  // Our changes affect the displayed style, but they don't "stick"
  // as far as another call to SetWindowLong(). Thus we keep track
  // of the style independently using the atom.
  IF Normal THEN
    begin
      REPEAT
        GlobalDeleteAtom(GlobalFindAtom(cIcoAtomName));
      UNTIL GlobalFindAtom(cIcoAtomName) = 0;
    end
  ELSE GlobalAddAtom(cIcoAtomName);
  H := lvHan;
  Styles.StyleOld := GetWindowLong(H, GWL_STYLE);
  Styles.StyleNew := Styles.StyleOld AND NOT LVS_TYPEMASK;
  Styles.StyleOld := Styles.StyleNew;
  IF Normal THEN
    begin
      Styles.StyleNew := Styles.StyleNew OR LVS_ICON;
      Styles.StyleOld := Styles.StyleOld OR LVS_SMALLICON;
    end
  ELSE
    begin
      Styles.StyleNew := Styles.StyleNew OR LVS_SMALLICON;
      Styles.StyleOld := Styles.StyleOld OR LVS_ICON;
    end;
  SetWindowLong(H, GWL_STYLE, Styles.StyleNew);
  SendMessage(H, WM_STYLECHANGED, GWL_STYLE, Integer(@Styles));
  ShowWindow(H, SW_SHOWNORMAL);
  fNormIcon := Normal;
end;

procedure TLayout.ToDesktop(TC : TCanvas);
// IF the layout's fPredef field reflects a specific predefined
// layout, then re-apply that layout. Do so using a temporary TLayout
// that holds all the icons currently on the desktop. If the fPredef
// field is -1, explicitly set icon positions.
VAR TMP : TLayout;
begin
  IF fPredef < 0 THEN
    ToDesktop_internal(TC)
  ELSE
    begin
      TMP := TLayout.Create;
      try
        TMP.FromDesktop;
        TMP.fNormIcon := Self.fNormIcon;
        SetListViewStyleTo(Self.fNormIcon);
        TMP.SetOrder(fList);
        TMP.SetLayout(fPredef, TC);
        TMP.ToDesktop(TC);
      finally
        TMP.Free;
      end;
    end;
end;

function TLayout.GetHProc : THandle;
// NT-family only. Get the handle of the process owning the
// listview. We need the handle to allocate memory within
// the process
type
  EnumProcessesType = function(
    hModules : Pointer; cb : DWORD; VAR
    cbNeeded : DWORD) : Bool; stdcall;
VAR
  PID             : DWORD;
  EnumProcesses   : EnumProcessesType;
  LibH            : hModule;
  hMods           : ARRAY OF THandle;
  cb, cbneeded, N : DWORD;
begin
  Result := 0;
  GetWindowThreadProcessID(LVHan, @PID);
  LibH := LoadLibrary('psapi.dll');
  IF LibH <= 32 THEN
    Raise(Exception.Create('PSAPI.DLL: ' +
      SysErrorMessage(GetLastError)));
  @EnumProcesses := GetProcAddress(LibH,
    'EnumProcesses');
  IF @EnumProcesses = nil THEN
     Raise(Exception.Create('Function EnumProcesses '+
       'not found in PSAPI.DLL '));
  SetLength(hMods, 1024);
  cb := 1024; cbneeded := 0;
  IF NOT EnumProcesses(hMods, cb, cbNeeded) THEN
    Exit;
  SetLength(hMods, cbNeeded DIV SizeOf(THandle));
  IF NOT EnumProcesses(hMods, cb, cbNeeded) THEN
    Exit;
  FOR N := 0 TO Length(hMods)-1 DO
    IF hMods[N] = PID THEN
      begin
        Result := OpenProcess(PROCESS_ALL_ACCESS, False, hMods[N]);
        Break;
      end;
end;

function TLayout.AllocShared : Boolean;
// Allocate shared memory areas for use within the program.
// Under NT-family, use VirtualAllocEx. Under Win9x, create
// a memory-mapped file (memory mappings are visible to all
// processes under Win9x)
begin
  Result := False;
  IF UnderNT THEN
    begin
      KernelH := LoadLibrary('Kernel32.dll');
      IF KernelH <= 32 THEN
        Raise(Exception.Create('KERNEL32.DLL: ' +
          SysErrorMessage(GetLastError)));
      // We can get away with freeing KernelH, because we KNOW its
      // reference count will not go to 0.
      FreeLibrary(KernelH);
      @vVirtualAllocEx := GetProcAddress(KernelH,
        'VirtualAllocEx');
      IF @vVirtualAllocEx = nil THEN
         Raise(Exception.Create('Function VirtualAllocEx '+
           'not found in KERNEL32.DLL '));
      @vVirtualFreeEx := GetProcAddress(KernelH,
        'VirtualFreeEx');
      IF @vVirtualFreeEx = nil THEN
         Raise(Exception.Create('Function VirtualFreeEx '+
           'not found in KERNEL32.DLL '));
      R_buf := vVirtualAllocEx(lvHProc, nil, MAX_PATH,
        MEM_COMMIT, PAGE_READWRITE);
      IF R_buf = nil THEN Exit;
      R_pt := vVirtualAllocEx(lvHProc, nil, SizeOf(TPoint),
        MEM_COMMIT, PAGE_READWRITE);
      IF R_pt = nil THEN Exit;
      R_lvi := vVirtualAllocEx(lvHProc, nil, SizeOf(TLVItem),
        MEM_COMMIT, PAGE_READWRITE);
      IF R_lvi = nil THEN Exit;
    end
  ELSE
    begin
      Map9x := CreateFileMapping($FFFFFFFF, NIL, PAGE_READWRITE,
        0, SizeOf(TShared9x), cMappingName);
      IF Map9x = 0 THEN Exit;
      Share9x := MapViewOfFile(Map9x, FILE_MAP_ALL_ACCESS, 0, 0, 0);
      IF Share9x = nil THEN Exit;
      R_buf := @Share9x^.buff;
      R_lvi := @Share9x^.LVI;
      R_pt  := @Share9x^.Pt;
    end;
  Result := True;
end;

procedure TLayout.ToShared;
// Transfer data from the local buffers to the shared
// memory area
VAR Actual : Cardinal;
begin
  IF UnderNT THEN
    begin
      WriteProcessMemory(lvHProc, R_buf, @L_buf, MAX_PATH, Actual);
      WriteProcessMemory(lvHProc, R_pt,  @L_pt,  PSiz,     Actual);
      WriteProcessMemory(lvHProc, R_lvi, @L_lvi, LSiz,     Actual);
    end
  ELSE
    begin
      StrCopy(R_buf, @L_buf);
      R_pt^  := L_pt;
      R_lvi^ := L_lvi;
    end;
end;

procedure TLayout.FromShared;
// Transfer data to the local buffers from the shared
// memory area
VAR Actual : Cardinal;
begin
  IF UnderNT THEN
    begin
      ReadProcessMemory(lvHProc, R_buf, @L_buf, MAX_PATH, Actual);
      ReadProcessMemory(lvHProc, R_pt,  @L_pt,  Psiz,     Actual);
      ReadProcessMemory(lvHProc, R_lvi, @L_lvi, Lsiz,     Actual);
    end
  ELSE
    begin
      StrCopy(@L_buf, R_buf);
      L_pt := R_pt^;
      L_lvi := R_lvi^;
    end;
end;

procedure TLayout.DoneShared;
// Free the shared memory areas
begin
  IF UnderNT THEN
    begin
      vVirtualFreeEx(lvHProc, R_buf, 0, MEM_RELEASE);
      vVirtualFreeEx(lvHProc, R_pt,  0, MEM_RELEASE);
      vVirtualFreeEx(lvHProc, R_lvi, 0, MEM_RELEASE);
      FreeLibrary(KernelH);
    end
  ELSE
    begin
      UnMapViewOfFile(share9x);
      CloseHandle(Map9x);
    end;
end;

procedure TLayout.SetHig(const Value: Integer);
begin
  fHig := Value;
  IF fHig <> -1 THEN
    fName := '';
end;

procedure TLayout.SetName(const Value: String);
begin
  fName := Value;
  IF fName <> '' THEN
    begin
      fWid := -1;
      fHig := -1;
    end;
end;

procedure TLayout.SetWid(const Value: Integer);
begin
  fWid := Value;
  IF fWid <> -1 THEN
    fName := '';
end;

function TLayout.lvHan: HWnd;
begin
  Result := GetLVHan;
end;

function TLayout.StoreStr(const S: String;
  vList : TStringList): String;
// Return the string to represent a particular icon. Prefix it with
// from 1: to 5:, and append the appropriate extension, if necessary.
// This routine assumes that the list is being filled starting with
// the *last* icon, in the desktop's internal order. SO e.g. if there
// are links named My Computer.lnk in both desktop folders, the first
// will be marked 4: (link in user desktop). The second can't have
// that designation because it already exists, so it will be marked
// 2: (common desktop). Since both of those are used up, the last
// one must be a non-file item (marked 1:).
VAR Ext : String;

  function FileIn(const where, what : String; C : Char) : Boolean;
  begin
    Result := (FileExists(Format('%s\%s', [where, what])) OR
      DirectoryExists(Format('%s\%s', [where, what]))) AND
      (vList.IndexOf(Format('%s:%s', [C, what])) < 0)
  end;

  function LinkIn(const where, what : String; C : Char; VAR vExt : String) : Boolean;
  VAR N : Integer;
  begin
    Result := False;
    FOR N := 0 TO ShortExt.Count-1 DO
      IF FileExists(Format('%s\%s%s', [where, what, ShortExt[N]])) AND
        (vList.IndexOf(Format('%s:%s%s', [C, what, ShortExt[N]])) < 0)THEN
        begin
          Result := True;
          vExt := ShortExt[N];
          Break;
        end;
  end;

begin
  IF LinkIn(DesktopFolder, S, '4', Ext) AND
    (vList.IndexOf('4:'+S) < 0) THEN
    Result := '4:'+S+Ext
  ELSE IF FileIn(DesktopFolder, S, '5') THEN
    Result := '5:'+S
  ELSE IF LinkIn(CommonFolder, S, '2', Ext) AND
    (vList.IndexOf('2:'+S) < 0) THEN
    Result := '2:'+S+Ext
  ELSE IF FileIn(CommonFolder, S, '3') THEN
    Result := '3:'+S
  ELSE Result := '1:'+S;
end;

procedure TLayout.MakeLayout(TS : TStringList; TC : TCanvas;
  vLayout : Integer);
// Build a string list containing the coordinates for the specified
// predefined layout. The strings have the form 0000,0000 and are
// used only in debugging. The actual data is stored by shoehorning
// TSmallPoint values (4 bytes) into the Objects items (4 bytes).
VAR
  XSize   : Integer; // grid size in X direction
  YSize   : Integer; // grid size in Y direction
  XScreen : Integer; // width of desktop listview
  YScreen : Integer; // height of desktop listview
  XLeft   : Integer; // starting position for left-to-right
  YTop    : Integer; // starting position for top-to-bottom
  XRight  : Integer; // starting position for right-to-left
  YBot    : Integer; // starting position for bottom-to-top
  cX      : Integer; // current X-value
  cY      : Integer; // current Y-value
  XAdj    : Integer; // adjustment for initial X
  Cols    : Integer; // number of columns in grid
  Rows    : Integer; // number of rows in grid
  R       : TRect;
  IM      : ICONMETRICS;

  procedure AddPos(X, Y : Integer);
  begin
    TS.AddObject(Format('%.04d,%.04d', [
      X, Y]), Pointer(PointToSmallPoint(Point(X,Y))));
  end;

  procedure TBSpiral;
  VAR
    ColLft  : Integer; // left column (changes as we spiral inward)
    ColRgt  : Integer; // right column (changes as we spiral inward)
    RowTop  : Integer; // top row (changes as we spiral inward)
    RowBot  : Integer; // bottom row (changes as we spiral inward)
    XDelt   : Integer; // -1, 0, or 1 - X direction to next item
    YDelt   : Integer; // -1, 0, or 1 - Y direction to next item
  begin
    ColLft := 0;
    ColRgt := Cols-1;
    RowTop := 0;
    RowBot := Rows-1;
    cX     := 0;
    cY     := 0;
    XDelt  := 0;
    YDelt  := 1; // Go down to start
    REPEAT
      AddPos(XLeft+cX*XSize, YTop+cY*YSize);
      IF cX+XDelt > ColRgt THEN
        begin
          // Turn from right to up
          XDelt := 0;
          YDelt := -1;
          Dec(RowBot);
        end
      ELSE IF cY+YDelt > RowBot THEN
        begin
          // Turn from down to right
          XDelt := 1;
          YDelt := 0;
          Inc(ColLft);
        end
      ELSE IF cX+XDelt < ColLft THEN
        begin
          // Turn from left to down
          XDelt := 0;
          YDelt := 1;
          Inc(RowTop);
        end
      ELSE IF cY+YDelt < RowTop THEN
        begin
          // Turn from up to left
          XDelt := -1;
          YDelt := 0;
          Dec(ColRgt);
        end;
      Inc(cX, XDelt);
      Inc(cY, YDelt);
    UNTIL TS.Count = Cols*Rows;
  end;

  procedure LRSpiral;
  VAR
    ColLft  : Integer; // left column (changes as we spiral inward)
    ColRgt  : Integer; // right column (changes as we spiral inward)
    RowTop  : Integer; // top row (changes as we spiral inward)
    RowBot  : Integer; // bottom row (changes as we spiral inward)
    XDelt   : Integer; // -1, 0, or 1 - X direction to next item
    YDelt   : Integer; // -1, 0, or 1 - Y direction to next item
  begin
    ColLft := 0;
    ColRgt := Cols-1;
    RowTop := 0;
    RowBot := Rows-1;
    cX     := 0;
    cY     := 0;
    XDelt  := 1; // Go right to start
    YDelt  := 0;
    REPEAT
      AddPos(XLeft+cX*XSize, YTop+cY*YSize);
      IF cX+XDelt > ColRgt THEN
        begin
          // Turn from right to down
          XDelt := 0;
          YDelt := 1;
          Inc(RowTop);
        end
      ELSE IF cY+YDelt > RowBot THEN
        begin
          // Turn from down to left
          XDelt := -1;
          YDelt := 0;
          Dec(ColRgt);
        end
      ELSE IF cX+XDelt < ColLft THEN
        begin
          // Turn from left to up
          XDelt := 0;
          YDelt := -1;
          Dec(RowBot);
        end
      ELSE IF cY+YDelt < RowTop THEN
        begin
          // Turn from up to right
          XDelt := 1;
          YDelt := 0;
          Inc(ColLft);
        end;
      Inc(cX, XDelt);
      Inc(cY, YDelt);
    UNTIL TS.Count = Cols*Rows;
  end;

begin
  IF IconsAreNormal THEN
    begin
      // Set the XSize and YSize to the icon spacing values
      FillChar(IM, SizeOf(IM), 0);
      IM.cbSize := SizeOf(IM);
      SystemParametersInfo(SPI_GETICONMETRICS, SizeOf(IM), @IM, 0);
      XSize := IM.iHorzSpacing;
      YSize := IM.iVertSpacing;
      WITH TRegIniFile.Create('Control Panel\desktop\WindowMetrics') DO
      try
        XAdj := (XSize - ReadInteger('', 'Shell Icon Size', 32)) DIV 2;
      finally
        Free;
      end;
    end
  ELSE
    begin
      // Set the XSize to the widest item, YSize to small icon height
      XAdj  := 4;
      YSize := GetSystemMetrics(SM_CYSMICON) + 2;
      XSize := Longest(TC) + YSize;
    end;
  GetWindowRect(lvHan, R);
  // Get the width and height of the desktop listview
  XScreen := R.Right-R.Left;
  YScreen := R.Bottom-R.Top;
  // Calculate how manu columns and rows will fit
  Cols := XScreen DIV XSize;
  Rows := YScreen DIV YSize;
  // Adjust the icon spacing to take advantage of the full area
  XSize := XScreen DIV Cols;
  YSize := YScreen DIV Rows;
  YTop   := 2;
  YBot   := YScreen - YSize - 2;
  XLeft  := XAdj;
  XRight := XScreen - XSize + XAdj;
  CASE vLayout OF
    // The first 8 layouts are easy
    cTbLr : begin
      FOR cX := 0 TO Cols-1 DO
        FOR cY := 0 TO Rows-1 DO
          AddPos(XLeft + cX*XSize, YTop + cY*YSize);
    end;
    cTbRl : begin
      FOR cX := 0 TO Cols-1 DO
        FOR cY := 0 TO Rows-1 DO
          AddPos(XRight - cX*XSize, YTop + cY*YSize);
    end;
    cBtLr : begin
      FOR cX := 0 TO Cols-1 DO
        FOR cY := 0 TO Rows-1 DO
          AddPos(XLeft + cX*XSize, YBot - cY*YSize);
    end;
    cBtRl : begin
      FOR cX := 0 TO Cols-1 DO
        FOR cY := 0 TO Rows-1 DO
          AddPos(XRight - cX*XSize, YBot - cY*YSize);
    end;
    cLrTb : begin
      FOR cY := 0 TO Rows-1 DO
        FOR cX := 0 TO Cols-1 DO
          AddPos(XLeft + cX*XSize, YTop + cY*YSize);
    end;
    cRlTb : begin
      FOR cY := 0 TO Rows-1 DO
        FOR cX := 0 TO Cols-1 DO
          AddPos(XRight - cX*XSize, YTop + cY*YSize);
    end;
    cLrBt : begin
      FOR cY := 0 TO Rows-1 DO
        FOR cX := 0 TO Cols-1 DO
          AddPos(XLeft + cX*XSize, YBot - cY*YSize);
    end;
    cRlBt : begin
      FOR cY := 0 TO Rows-1 DO
        FOR cX := 0 TO Cols-1 DO
          AddPos(XRight - cX*XSize, YBot - cY*YSize);
    end;
    cLrSp : LRSpiral;
    cTbSp : TBSpiral;
    cTbIn : begin
      FOR cX := 0 TO (Cols DIV 2)-1 DO
        begin
          // Add a column down the left
          FOR cY := 0 TO Rows-1 DO
            AddPos(XLeft + cX*XSize, YTop + cY*YSize);
          // Add a column down the right
          FOR cY := 0 TO Rows-1 DO
            AddPos(XRight - cX*XSize, YTop + cY*YSize);
        end;
      // Add the middle column, if any
      IF Odd(Cols) THEN
        begin
          cX := Cols DIV 2;
          FOR cY := 0 TO Rows-1 DO
            AddPos(XLeft + cX*XSize, YTop + cY*YSize);
        end;
    end;
    cLrIn : begin
      FOR cY := 0 TO (Rows DIV 2)-1 DO
        begin
          // Add a row across the top
          FOR cX := 0 TO Cols-1 DO
            AddPos(XLeft + cX*XSize, YTop + cY*YSize);
          // Add a row across the bottom
          FOR cX := 0 TO Cols-1 DO
            AddPos(XLeft + cX*XSize, YBot - cY*YSize);
        end;
      // Add the middle row, if any
      IF Odd(Rows) THEN
        begin
          cY := Rows DIV 2;
          FOR cX := 0 TO Cols-1 DO
            AddPos(XLeft + cX*XSize, YTop + cY*YSize);
        end;
    end;
  END;
end;

procedure TLayout.ToDesktop_internal(TC : TCanvas);
// Set the icons in the precise locations specified in the layout.
// Note any that have no location by adding to the fRemains list.
VAR
  N     : Integer;
  Count : Integer;
  H     : HWnd;
  TS    : TStringList;
  lvR   : TRect;
  XSize : Integer;
  YSize : Integer;
  XDelt : Integer;
  YDelt : Integer;

  procedure GetSizes(IsNormal : Boolean);
  VAR IM : ICONMETRICS;
  begin
    IF IsNormal THEN
      begin
        // Set the XSize and YSize to the icon spacing values
        FillChar(IM, SizeOf(IM), 0);
        IM.cbSize := SizeOf(IM);
        SystemParametersInfo(SPI_GETICONMETRICS, SizeOf(IM), @IM, 0);
        XSize := IM.iHorzSpacing;
        YSize := IM.iVertSpacing;
      end
    ELSE
      begin
        // Set the XSize to the widest item, YSize to small icon height
        YSize := GetSystemMetrics(SM_CYSMICON) + 2;
        XSize := Longest(TC) + YSize;
      end;
    XDelt := 0;
    YDelt := 0;
  end;

  function ForceIntoView(VAR P : TPoint) : Boolean;
  begin
    Result := False;
    IF P.X < lvR.Left THEN
      begin
        P.X := lvR.Left + XDelt;
        Result := True;
      end;
    IF P.Y < lvR.Top THEN
      begin
        P.Y := lvR.Top + YDelt;
        Result := True;
      end;
    IF P.X > lvR.Right - XSize THEN
      begin
        P.X := lvR.Right - XSize - XDelt;
        Result := True;
      end;
    IF P.Y > lvR.Bottom - YSize THEN
      begin
        P.Y := lvR.Bottom - YSize - YDelt;
        Result := True;
      end;
  end;

  procedure SetOneIcon(N : Integer);
  VAR
    Idx : Integer;
    sN : String;
  begin
    WITH L_lvi DO
      begin
        mask       := LVIF_TEXT;
        iItem      := N;
        iSubItem   := 0;
        state      := 0;
        statemask  := 0;
        pszText    := R_buf;
        cchTextMax := MAX_PATH;
        iImage     := 0;
      end;
    ToShared;
    SendMessage(H, LVM_GETITEM, 0, Integer(R_lvi));
    FromShared;
    sN := StoreStr(L_buf, TS);
    TS.Add(sN);
    Idx := fList.IndexOf(sN);
    IF Idx = -1 THEN
      begin
        // Make sure the desktop hasn't flung this item offscreen
        // due to normal/small change
        ToShared;
        SendMessage(H, LVM_GETITEMPOSITION, N, Integer(R_Pt));
        FromShared;
        IF ForceIntoView(L_Pt) THEN
          begin
            ToShared;
            SendMessage(H, LVM_SETITEMPOSITION32, N, Integer(R_pt));
            FromShared;
            Inc(XDelt, 16);
            Inc(YDelt, 16);
          end;
        FRemains.Add(L_buf);
      end
    ELSE
      begin
        L_pt := PointerToPoint(fList.Objects[Idx]);
        ToShared;
        SendMessage(H, LVM_SETITEMPOSITION32, N, Integer(R_pt));
        FromShared;
      end;
  end;

begin
  H := lvHan;
  GetWindowRect(H, lvR);
  GetSizes(fNormIcon);
  try
    SetListviewStyleTo(fNormIcon);
    fRemains.Clear;
    IF UnderNT THEN lvHProc := GetHProc;
    IF NOT AllocShared THEN Exit;
    Count := SendMessage(H, LVM_GETITEMCOUNT, 0, 0);
    try
      TS := TStringList.Create;
      try
        FOR N := Count-1 DOWNTO 0 DO
          SetOneIcon(N);
      finally
        TS.Free;
      end;
    finally
      DoneShared;
      SendMessage(H, LVM_REDRAWITEMS, 0, Count-1);
    end;
  finally
    IF lvHProc <> 0 THEN CloseHandle(lvHProc);
  end;
end;


{ TLayoutList }

constructor TLayoutList.Create;
begin
  Inherited Create;
  fNamed := TStringList.Create;
  fResol := TStringList.Create;
  fOrder := TLayout.Create;
end;

destructor TLayoutList.Destroy;
VAR N : Integer;
begin
  FOR N := 0 TO fNamed.Count-1 DO
    TLayout(fNamed.Objects[N]).Free;
  FOR N := 0 TO fResol.Count-1 DO
    TLayout(fResol.Objects[N]).Free;
  fNamed.Free;
  fResol.Free;
  fOrder.Free;
  inherited;
end;

procedure TLayoutList.SaveToStream(Stream: TStream);
VAR N : Integer;
begin
  Stream.WriteBuffer(Signature, 4);
  Stream.WriteBuffer(DatVersion, 4);
  fOrder.SaveToStream(Stream);
  N := fNamed.Count + fResol.Count;
  Stream.Write(N, 4);
  FOR N := 0 TO fNamed.Count-1 DO
    TLayout(fNamed.Objects[N]).SaveToStream(Stream);
  FOR N := 0 TO fResol.Count-1 DO
    TLayout(fResol.Objects[N]).SaveToStream(Stream);
end;

procedure TLayoutList.LoadFromStream(Stream: TStream);
VAR
  N     : Integer;
  D     : DWord;
  Count : DWord;
  TL    : TLayout;
begin
  Stream.ReadBuffer(D, 4);
  IF D <> Signature THEN
    Raise(Exception.Create('sig'));
  Stream.ReadBuffer(D, 4);
  IF D <> DatVersion THEN
    Raise(Exception.Create('ver'));
  fOrder.LoadFromStream(Stream);
  Stream.ReadBuffer(Count, 4);
  FOR N := 0 TO Count-1 DO
    begin
      TL := TLayout.Create;
      TL.LoadFromStream(Stream);
      AddLayout(TL);
    end;
end;

procedure TLayoutList.SaveToFile(const S: String);
VAR Stream : TStream;
begin
  Stream := TFileStream.Create(S, fmCreate OR fmShareDenyWrite);
  try
    SaveToStream(STream);
  finally
    Stream.Free;
  end;
end;

procedure TLayoutList.LoadFromFile(const S: String);
VAR Stream : TStream;
begin
  Stream := TFileStream.Create(S, fmOpenRead OR fmShareDenyWrite);
  try
    LoadFromStream(STream);
  finally
    Stream.Free;
  end;
end;

procedure TLayoutList.AddLayout(vTL: TLayout);
VAR TL : TLayout;
begin
  WITH vTL DO
    IF fName <> '' THEN
      begin
        TL := LayoutByName(fName);
        IF TL <> nil THEN
          DelLayout(fName);
        fNamed.AddObject(fName, vTL)
      end
    ELSE IF (fWid >= 0) AND (fHig >= 0) THEN
      begin
        TL := LayoutByRes(fWid, fHig);
        IF TL <> nil THEN
          DelLayout(fWid, fHig);
        fResol.AddObject(Format('%dx%d', [fWid, fHig]), vTL)
      end
    ELSE
      Raise(Exception.Create('Invalid TLayout'));
end;

procedure TLayoutList.DelLayout(const Name: String);
VAR Idx : Integer;
begin
  Idx := fNamed.IndexOf(Name);
  IF Idx <> -1 THEN
    fNamed.Delete(Idx);
end;

procedure TLayoutList.DelLayout(vWid, vHig: Integer);
VAR Idx : Integer;
begin
  Idx := fResol.IndexOf(Format('%dx%d', [vWid, vHig]));
  IF Idx <> -1 THEN
    fResol.Delete(Idx);
end;

function TLayoutList.LayoutByName(const Name: String): TLayout;
VAR Idx : Integer;
begin
  Idx := fNamed.IndexOf(Name);
  IF Idx = -1 THEN Result := nil
  ELSE Result := TLayout(fNamed.Objects[Idx]);
end;

function TLayoutList.LayoutByRes(vWid, vHig: Integer): TLayout;
VAR Idx : Integer;
begin
  Idx := fResol.IndexOf(Format('%dx%d', [vWid, vHig]));
  IF Idx = -1 THEN Result := nil
  ELSE Result := TLayout(fResol.Objects[Idx]);
end;

end.
